/******************************************************************************
*******************************************************************************
*******************************************************************************

    rebler
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: rebler.cpp,v 1.4 2001/08/06 13:59:36 monkeyiq Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/


#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <memory>
#include <iomanip.h>
#include <popt.h>

using namespace std;

// libavi
#include <avifile.h>
#include <fourcc.h>
#include <except.h>
#include <videoencoder.h>
#include <creators.h>
using namespace Creators;


const string PROGRAM_NAME = "rebler";

void usage(poptContext optCon, int exitcode, char *error, char *addl)
{
    poptPrintUsage(optCon, stderr, 0);
    if (error) fprintf(stderr, "%s: %s0", error, addl);
    exit(exitcode);
}

const string tostr( ostringstream& oss )  { return oss.rdbuf()->str(); }
int atexit(void (*function)(void))
{
    return ::atexit( function );
}

const string tostr( fourcc_t v )
{
    ostringstream ss;

    ss << (char)((v)       & 0xFF );
    ss << (char)((v >>  8) & 0xFF );
    ss << (char)((v >> 16) & 0xFF );
    ss << (char)((v >> 24) & 0xFF );
    return tostr(ss);
}


bool CanEncode( const CodecInfo* ci )
{
    return ((ci->direction & CodecInfo::Encode) || (ci->direction & CodecInfo::Both));
}






class Rebler
{
    string InputVOBFilename;
    string InputAC3Filename;
    string OutputAVIFilename;
    string OutputFourCC;
    
    double IVideoFrameRate;
    int InputVideoStreamNumber;
    int InputAudioStreamNumber;
    int TotalInputVideoFrameCount;
    int TotalInputAudioStreamCount;

    int IVideoWidth;
    int IVideoHeight;

    ifstream              InputAC3Stream;
//    ifstream&             InputPGMPipe;
    _IO_istream_withassign& InputPGMPipe;
    
    IAviWriteFile*        OutputStream;
    IAviVideoWriteStream* OutputVStream;
    IAviWriteStream*      OutputAStream;

    unsigned char* SpareFrameBufferRow;
    unsigned char* FrameBuffer;
    unsigned char** OutputRows;

    unsigned char* AudioBuffer;
    long AudioToReadPerFrame;
    
    fourcc_t      OutputCodecID;
    int           OutputBitRate;

    bool VeritcalFlipOutputVideo;
    

    BitmapInfo  OutputBitmapInfo;
    CImage*     OutputImage;

    bool MoreIVideo;
    bool MoreIAudio;

    
public:

    Rebler()
        :
        InputPGMPipe( cin ),
        OutputStream(0),
        OutputAStream(0),
        OutputVStream(0),
        FrameBuffer(0),
        SpareFrameBufferRow(0),
        OutputRows(0),
        OutputCodecID(fccDIV3), //DIV6
        OutputBitRate(1800),
        AudioBuffer(0),
        AudioToReadPerFrame(0),
        VeritcalFlipOutputVideo(false),
        TotalInputVideoFrameCount(-1),
        MoreIVideo(true),
        MoreIAudio(true)
        {
            memset( &OutputBitmapInfo, 0, sizeof(BitmapInfo));
            
        }

    virtual ~Rebler()
        {

            delete [] SpareFrameBufferRow;
            delete [] FrameBuffer;
            delete [] OutputRows;
            delete [] AudioBuffer;
        }


    void setOutputFourCC( const string& v )
        {
            if( v.length() >= 4 )
            {
                OutputFourCC = v;
                OutputCodecID = mmioFOURCC( v[0], v[1], v[2], v[3] );
                cout << "Set fourCC to :" << v << endl;
            }
        }
    

    
    void setVeritcalFlipOutputVideo( bool v )
        {
            VeritcalFlipOutputVideo = v;
        }
    

    
    void setInputVOBFilename( const string& v )
        {
            InputVOBFilename = v;
        }

    void setInputAC3Filename( const char* v )
        {
            if( v ) InputAC3Filename = v;
        }
    
    

    void setOutputAVIFilename( const string& v )
        {
            OutputAVIFilename = v;
        }

    void setInputVideoStreamNumber( int v )
        {
            InputVideoStreamNumber = v;
        }

    void setInputAudioStreamNumber( int v )
        {
            InputAudioStreamNumber = v;
        }

    void setOutputBitRate( int v )
        {
            OutputBitRate = v;
        }

    class CodecDataException : public exception
    {
        string w;
    public:
        CodecDataException( string s ) : w(s)
            {
            }

        const char* what()
            {
                return w.c_str();
            }
        
    };
    

    const CodecInfo* getCodecInfo()
        {
            return CodecInfo::match(OutputCodecID, CodecInfo::Video, 0);
        }
    
    
    void setCodecData( const string& k, const string& v )
        {
            const CodecInfo* ci = getCodecInfo();
            
            if(ci && CanEncode(ci))
            {
                cout << "Setting k: " << k << " to value:" << v << endl;
	            SetCodecAttr(*ci, k.c_str(), v.c_str() );
            }
            else
            {
                ostringstream ss;
                ss << "Can not set k:" << k << " to value:" << v << endl;
                throw CodecDataException( tostr(ss) );
            }

        }

    void setCodecData( const string& k, long v )
        {
            ostringstream ss;
            ss << v;
            setCodecData( k, tostr(ss) );
        }

    void setCodecData( const string& kv )
        {
            string k;
            string v;

            istringstream ss(kv);
            getline( ss, k, '=' );
            getline( ss, v );

            cerr << "k:" << k << endl;
            cerr << "v:" << v << endl;

            setCodecData( k, v );
        }
    
    
    
    void dumpCodec( const CodecInfo* ci )
        {
            cerr << "-------------------------------------------" << endl;
            cerr << "Name       : " << ci->text << endl;
            cerr << "About      : " << ci->about << endl;
            cerr << "Win32DLL   : " << ci->dll << endl;
//                cerr << "GUID       : " << ci->guid << endl;
            cerr << "module name: " << ci->modulename << endl;
            cerr << "Fourcc     : " << tostr(ci->fourcc) << endl;
            cerr << "Also handles fourcc :";

            for( vector<fourcc_t>::const_iterator fi = ci->fourcc_array.begin();
                 fi != ci->fourcc_array.end(); ++fi )
            {
                cerr << " " << tostr(*fi);
            }
            cerr << endl;
            
            cerr << "kind       : ";
            switch( ci->kind )
            {
            case CodecInfo::Source: cerr << "Source "; break;
            case CodecInfo::Plugin: cerr << "Plugin "; break;
            case CodecInfo::Win32:  cerr << "Win32  "; break;
            case CodecInfo::DShow_Dec: cerr << "DShow_Dec"; break;
            case CodecInfo::DShow_Enc: cerr << "DShow_Enc"; break;
            default:
                cerr << "Unknown";
                break;
            }
            cerr << endl;
            
            if( ci->media == CodecInfo::Audio )
                cerr << "Audio codec" << endl;
            else if( ci->media == CodecInfo::Video )
                cerr << "Video codec" << endl;
            
            if( ci->direction == CodecInfo::Both || ci->direction == CodecInfo::Encode )
                cerr << "Can encode" << endl;
            if( ci->direction == CodecInfo::Both || ci->direction == CodecInfo::Decode )
                cerr << "Can decode" << endl;

            for( std::vector<AttributeInfo>::const_iterator iter =
                     ci->encoder_info.begin();
                 iter != ci->encoder_info.end();
                 iter++)
            {
                cerr << "  " << "name  : " << iter->name << endl;
                cerr << "    " << "about : " << iter->about << endl;
                cerr << "    " << "kind  : ";
                switch(iter->kind)
                {
                case AttributeInfo::Integer:
                    cerr << "Integer, min:" << iter->i_min << " max:" << iter->i_max;
                    break;
                case AttributeInfo::String:  cerr << "String";  break;
                case AttributeInfo::Select:
                {
                    cerr << "Select from: ";
                    for( std::vector<std::string>::const_iterator i =
                             iter->options.begin();
                         i != iter->options.end();
                         i++ )
                    {
                        cerr << *i << " ";
                    }
                    break;
                }
                default:
                    cerr << "unknown!" << endl;
                    break;
                }
            cerr << endl;
                
            }
            
            
            
        }
    


    
    void dumpCodecs()
        {
            typedef std::vector<CodecInfo> CodecInfo_t;

            for( CodecInfo_t::iterator iter = video_codecs.begin();
                 iter != video_codecs.end();
                 iter++)
            {
                dumpCodec(iter);
            }
        }
    

    /*
     * Keeping this around in case libmpeg2 doesn't support flipping.
     * I could make it faster with libmpeg3 using the OutputBuffer and
     * changing it to be preflipped.
     */
    void VerticalFlip()
        {
//             const int h = IVideoHeight;
//             const int w = IVideoWidth;
//             const int obytespp = 3;
//             const int woffset = w*obytespp;
            
//             unsigned char* tmp = SpareFrameBufferRow;
            
//             for( int i=0; i < (h/2); ++i )
//             {
//                 unsigned char* toprow = &FrameBuffer[(i)  *woffset];
//                 unsigned char* botrow = &FrameBuffer[(h-i-1)*woffset];
                
//                 copy( toprow, toprow+woffset, tmp );
//                 copy( botrow, botrow+woffset, toprow );
//                 copy( tmp, tmp+woffset, botrow );
//             }
        }

    void parseIVideoHeader()
        {
            string s;
            getline( InputPGMPipe, s );
            getline( InputPGMPipe, s );

            InputPGMPipe >> IVideoWidth;
            InputPGMPipe >> IVideoHeight;

            IVideoHeight /= 3;
            IVideoHeight *= 2;
            
            while( s != "255" )
            {
                getline( InputPGMPipe, s );
            }
            
        }


    void readIVideoFrame( bool includesHeader = true )
        {
            if( includesHeader )
            {
                parseIVideoHeader();
            }
            const int iw    = IVideoWidth;
            const int ih    = IVideoHeight;
            const int plane = iw*ih;
            const int imagesz = (int)(iw*ih*1.5);
            const int hw = iw / 2;
            
            InputPGMPipe.read( &FrameBuffer[             0 ], plane );
            for (int i = 0; i < ih >> 1; i++)
            {
                InputPGMPipe.read( &FrameBuffer[ plane + plane/4 + i*hw ], hw );
                InputPGMPipe.read( &FrameBuffer[ plane + i*hw ], hw );
            }
        
            if( !InputPGMPipe.gcount() || !InputPGMPipe )
            {
                MoreIVideo = false;
            }
        }
    
    
    virtual bool ExposeFrame()
        {
            const int ix = 0;
            const int iy = 0;
            const int iw = IVideoWidth;
            const int ih = IVideoHeight;
            const int ow = iw;
            const int oh = ih;
            const int obytespp = 3;
            static int FramesExposed = 0;


            readIVideoFrame();
            
            

//             ostringstream ss;
//             ss << "out-" << InputAudioStreamNumber << "-" << setw(5) << setfill('0')
//                << FramesExposed;
//            cout << "Making frame:" << tostr(ss) << endl;

            if( VeritcalFlipOutputVideo )
            {
//                cerr << "flipping" << endl;
                VerticalFlip();
            }
            

            
            OutputVStream->AddFrame( OutputImage );

            ++FramesExposed;
            if( FramesExposed % 24 == 0 ) 
                cout << "Exposed " << FramesExposed << " frames" << endl;

        
            
//             ofstream os( tostr(ss).c_str() );
//             os.write( FrameBuffer, ow * oh * obytespp );

            return true;
        }

    

    inline void initOutputRows()
        {
            const int h = IVideoHeight;
            const int w = IVideoWidth;
            const int obytespp = 3;

            for( int i=0; i < h; ++i )
            {
                OutputRows[i] = &FrameBuffer[ i * w * obytespp ];
            }
            
        }



    void initIVideo()
        {
//            cout << "File opened:" << InputVOBFilename << endl << flush;

            parseIVideoHeader();

            
            const int iw = IVideoWidth;
            const int ih = IVideoHeight;
            const int obytespp = 3;

            cout << " Setting up OutputBitmapInfo, w:" << iw << " h:" << ih << endl;
            
//            BITMAPINFOHEADER bh;
            OutputBitmapInfo.biWidth=IVideoWidth;
            OutputBitmapInfo.biHeight=IVideoHeight;
            OutputBitmapInfo.biBitCount=16;
            OutputBitmapInfo.biCompression= fccYV12;
//            OutputBitmapInfo.biCompression= fccYUY2;
            
            OutputBitmapInfo.biSizeImage = 0;
            OutputBitmapInfo.biPlanes = 1;
            OutputBitmapInfo.biSize = sizeof(BITMAPINFOHEADER);
            OutputBitmapInfo.biXPelsPerMeter = OutputBitmapInfo.biYPelsPerMeter = 0;
            OutputBitmapInfo.biClrUsed = OutputBitmapInfo.biClrImportant = 256;

            cout << "Allocating buffer for one frame" << endl << flush;
            SpareFrameBufferRow = new (unsigned char)[ iw * obytespp + 4 ];
            FrameBuffer = new (unsigned char)[ iw * ih * obytespp + 4 ];
            OutputRows  = new (unsigned char*) [ih];
            initOutputRows();


            cout << "Input Width:" << iw << endl;
            cout << "Input Height:" << ih << endl;
            
            OutputImage = new CImage( &OutputBitmapInfo, FrameBuffer, false);

            cout << "Setting the output bitrate" << endl << flush;
            
            /*
             * Set bitrate.
             */
            try {
                setCodecData( "BitRate", OutputBitRate);
            }
            catch( CodecDataException& e )
            {
                cerr << "Error can not set attribute!" << endl;
                cerr << e.what() << endl;
                exit(1);
            }
            

            IVideoFrameRate = 24;//mpeg3_frame_rate(mpg, InputVideoStreamNumber);
            cout << "Adding the video stream to the output file:" << OutputStream << endl << flush;
            OutputVStream = OutputStream->AddVideoStream(
                OutputCodecID, &OutputBitmapInfo,
                (int) (1000000 / IVideoFrameRate));
            cout << "Added the video stream..." << endl;

            OutputVStream->SetKeyFrame(1);
            OutputVStream->Start();
            
            cout << "Done with video init" << endl << flush;

            readIVideoFrame( false );
        }

    
    

    void initInput()
        {
            initIVideo();


            /*
             * Audio init.
             */
            if( InputAC3Filename.length() )
            {
                
                InputAC3Stream.open(InputAC3Filename.c_str());

                const int rate = 384000;
//                const int rate = 48000;
                const int channels = 5;
            
                WAVEFORMATEX fmt;
                memset(&fmt, 0, sizeof(WAVEFORMATEX));
                fmt.wFormatTag = 0x2000;
                fmt.nChannels = channels;
                fmt.nSamplesPerSec = rate;
                fmt.nAvgBytesPerSec = rate * channels;
                fmt.nBlockAlign = channels;
                fmt.wBitsPerSample = 16;
                fmt.cbSize = sizeof(WAVEFORMATEX);
                
                try {
                    OutputAStream = OutputStream->AddStream(
                        AviStream::Audio, (char*)&fmt, sizeof(WAVEFORMATEX),
                        fmt.wFormatTag, fmt.nAvgBytesPerSec, fmt.nBlockAlign);

                }
                catch(FatalError & e) {
                    e.Print();
                    exit(1);
                }

//                double tmp = (channels * rate * 1.0)/ IVideoFrameRate;
//                double tmp = (rate * 1.0)/ IVideoFrameRate;
//                double tmp = 2000;
                double tmp = (rate / (8*IVideoFrameRate));
                AudioToReadPerFrame = (long)(tmp);
                AudioBuffer = new (unsigned char)[ AudioToReadPerFrame ];
            }
            
            cout << "Done with audio init" << endl << flush;
        }
    
    

    void operator()()
        {
            MoreIVideo = true;
            MoreIAudio = true;
            

//             InputPGMPipe.open( (char*)InputVOBFilename.c_str());
//             if( !InputPGMPipe )
//             {
//                 cerr << "Count not open pipe " << endl;
//                 exit(1);
//             }
            
            
            cout << "Creating the output AVI file at:" << OutputAVIFilename << endl << flush;
            OutputStream = CreateIAviWriteFile(OutputAVIFilename.c_str());
            initInput();
            
//             {
//                 unsigned char* x = new (unsigned char) [6000000];
                
//                 InputAC3Stream.read( x, 5912924 );
//                 cerr << "writing some adata:" << InputAC3Stream.gcount() << endl;
//                 OutputAStream->AddChunk(
//                     (const char*) x,
//                     (unsigned int)InputAC3Stream.gcount(), 0 );
//             }


            cout << "Total Frames:" << TotalInputVideoFrameCount << endl;

            
            
            const int IFPS = (int)IVideoFrameRate;
            
            for( long i=0; MoreIVideo && MoreIAudio; ++i )
            {
                if( MoreIVideo )
                {
                    ExposeFrame();
                }
                
                
 
//                if( !(i % IFPS))
                if( MoreIAudio && AudioBuffer && InputAC3Stream )
                {
                    InputAC3Stream.read( AudioBuffer, AudioToReadPerFrame );

//                    cerr << "writing some adata:" << InputAC3Stream.gcount() << endl;
                    OutputAStream->AddChunk(
                        (const char*) AudioBuffer,
                        (unsigned int)InputAC3Stream.gcount(), 0 );

                    if( !InputAC3Stream.gcount() )
                    {
                        MoreIAudio = false;
                    }
                }
                
            }

            delete OutputStream;
        }
    
    
};



void main( int argc, const char** argv )
{
    Rebler rebler;
    
    const char* OutputFourCC = 0;
    const char* InputVOBFilename = 0;
    const char* InputAC3Filename = 0;
    const char* OutputAVIFilename = 0;
    unsigned long InputVideoStreamNumber = 0;
    unsigned long InputAudioStreamNumber = 0;
    unsigned long OutputBitRate = 1800;
    unsigned long DumpCodecNames = 0;
    unsigned long VeritcalFlipOutputVideo = false;
    
    struct poptOption optionsTable[] = {
        { "output-fourcc", 0, POPT_ARG_STRING, &OutputFourCC, 0,
          "FourCC of video codec to use for encode", "" },

//         { "input-vob", 0, POPT_ARG_STRING, &InputVOBFilename, 0,
//           "Name of the VOB file to convert", "" },
        { "input-ac3", 'i', POPT_ARG_STRING, &InputAC3Filename, 0,
          "Name of the ac3 file to mux into output video", "" },

        { "output-avi", 'o', POPT_ARG_STRING, &OutputAVIFilename, 0,
          "Name of the AVI file to create", "" },

        { "input-video-stream", 'v', POPT_ARG_INT, &InputVideoStreamNumber, 0, "", 0 },
        { "input-audio-stream", 'a', POPT_ARG_INT, &InputAudioStreamNumber, 0, "", 0 },

        { "output-video-bitrate", 'b', POPT_ARG_INT, &OutputBitRate, 0, "", 0 },
        
        { "dump-codec-names", 'd', POPT_ARG_NONE, &DumpCodecNames, 0, "", 0 },
//         { "flip-output-video", 'f', POPT_ARG_NONE, &VeritcalFlipOutputVideo, 0,
//           "Flip each video frame in the virtical", 0 },
        
        POPT_AUTOHELP
        { NULL, 0, 0, NULL, 0 }
   };
    poptContext optCon;


    optCon = poptGetContext(PROGRAM_NAME.c_str(), argc, argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]");

    if (argc < 2) {
        poptPrintUsage(optCon, stderr, 0);
        exit(1);
    }
    
    
    /* Now do options processing, get portname */
    char c=-1;
    while ((c = poptGetNextOpt(optCon)) >= 0)
    {
//         switch (c) {
//         }
    }


    if( DumpCodecNames )
    {
        cerr << "Dumping the codecs that are available. " << endl;
        rebler.dumpCodecs();
    }
    

    try
    {

//        rebler.setInputVOBFilename( InputVOBFilename );
        if( InputAC3Filename )
            rebler.setInputAC3Filename( InputAC3Filename );
        rebler.setOutputAVIFilename( OutputAVIFilename );
        rebler.setInputVideoStreamNumber( InputVideoStreamNumber );
        rebler.setInputAudioStreamNumber( InputAudioStreamNumber );
        rebler.setOutputBitRate( OutputBitRate );
//        rebler.setVeritcalFlipOutputVideo( VeritcalFlipOutputVideo );
        
        if( OutputFourCC )
            rebler.setOutputFourCC( OutputFourCC );
        
    }
    catch( exception& e )
    {
        cerr << "e:" << e.what() << endl;
        exit(1);
    }


    const char* Attr;
    for( ; Attr = poptGetArg(optCon); )
    {
        cerr << "Found extra codec data:" << Attr << endl;
        rebler.setCodecData( Attr );
    }
    
    
    
    rebler();

    
    poptFreeContext(optCon);
    return 0;
}
